import logging

from aislib.misc_utils import get_logger as _get_logger


def get_logger(name: str) -> logging.Logger:
    logger = _get_logger(name=name)
    logger.propagate = False

    return logger



from pathlib import Path

import luigi
import pandas as pd
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class ParseLabelFile(luigi.Task):
    label_file_path = luigi.Parameter()
    output_folder = luigi.Parameter()

    def run(self) -> None:
        ensure_path_exists(path=self.output_path())

        df = pd.read_csv(filepath_or_buffer=str(self.label_file_path))
        df_parsed = _parse_label_df(df=df)

        df_parsed.to_csv(
            path_or_buf=self.output_path().with_suffix(".csv"), index=False
        )

    def output_path(self) -> Path:
        return Path(str(self.output_folder), "tabular/labels.csv")

    def output(self) -> luigi.LocalTarget:
        return luigi.LocalTarget(self.output_path())


def _parse_label_df(df: pd.DataFrame) -> pd.DataFrame:
    df_no_na = _remove_any_na_from_label_df(df=df)

    return df_no_na


def _remove_any_na_from_label_df(df: pd.DataFrame) -> pd.DataFrame:
    df_copy = df.copy()
    number_before = len(df_copy)
    df_no_na = df_copy.dropna(how="any")
    number_after = len(df_no_na)

    difference = number_before - number_after
    logger.info("Dropped %d rows that had NA in any column.", difference)

    return df_no_na



from pathlib import Path
from shutil import copyfile
from typing import List

import luigi
from aislib.misc_utils import ensure_path_exists
from luigi.util import requires
from plink_pipelines.make_dataset import (
    RenameOnFailureMixin,
    get_sample_generator_from_bed,
    _get_one_hot_encoded_generator,
    write_one_hot_outputs,
)

from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class Config(luigi.Task, RenameOnFailureMixin):
    output_folder = luigi.Parameter()

    @property
    def input_name(self):
        bed_files = [
            i for i in Path(str(self.raw_data_path)).iterdir() if i.suffix == ".bed"
        ]
        if len(bed_files) != 1:
            raise ValueError(
                f"Expected one .bed file in {self.raw_data_path}, but"
                f"found {bed_files}."
            )
        return str(bed_files[0])

    @property
    def file_name(self):
        raise NotImplementedError

    def output_target(self, file_name: str):
        output_path = Path(str(self.output_folder), file_name)

        return luigi.LocalTarget(str(output_path))

    def output(self):
        return self.output_target(self.file_name)


def _get_plink_inputs_from_folder(folder_path: Path) -> List[Path]:
    files = [i.with_suffix("") for i in folder_path.iterdir() if i.suffix == ".bed"]

    return files


class ExternalRawData(luigi.ExternalTask):
    raw_data_path = luigi.Parameter()

    @property
    def input_name(self):
        bed_files = [
            i for i in Path(str(self.raw_data_path)).iterdir() if i.suffix == ".bed"
        ]
        if len(bed_files) != 1:
            raise ValueError(
                f"Expected one .bed file in {self.raw_data_path}, but"
                f"found {bed_files}."
            )
        return str(bed_files[0])

    def output(self):
        return luigi.LocalTarget(str(self.input_name))


@requires(ExternalRawData)
class OneHotAutoSNPs(Config):
    """
    Generates one hot encodings from a individuals x SNPs file.
    """

    include_text = luigi.BoolParameter()
    output_folder = luigi.Parameter()
    output_format = luigi.Parameter()
    output_name = luigi.Parameter()
    file_name = str(output_name)

    def run(self):
        input_path = Path(self.input().path)
        assert input_path.suffix == ".bed"

        output_path = Path(self.output().path)
        ensure_path_exists(output_path, is_folder=True)

        chunk_generator = get_sample_generator_from_bed(
            bed_path=input_path, chunk_size=1000
        )
        sample_id_one_hot_array_generator = _get_one_hot_encoded_generator(
            chunked_sample_generator=chunk_generator
        )
        write_one_hot_outputs(
            id_array_generator=sample_id_one_hot_array_generator,
            output_folder=output_path,
            output_format=str(self.output_format),
            output_name=str(self.output_name),
        )

    def output_target(self, file_name: str):
        return f"{self.output_folder}/processed/encoded_outputs/{self.output_name}"

    def output(self):
        target = self.output_target(self.file_name)
        return luigi.LocalTarget(path=target)


@requires(OneHotAutoSNPs)
class FinalizeGenotypeParsing(luigi.Task):
    raw_data_path = luigi.Parameter()
    output_folder = luigi.Parameter()
    output_format = luigi.Parameter()
    output_name = luigi.Parameter()

    def run(self):
        raw_path = Path(str(self.raw_data_path))
        bim_files = [i for i in raw_path.iterdir() if i.suffix == ".bim"]
        assert len(bim_files) == 1
        bim_path = bim_files[0]

        output_path = Path(self.output()[0].path)
        ensure_path_exists(output_path, is_folder=False)

        copyfile(bim_path, output_path)

    def output(self):
        output_path = Path(
            str(self.output_folder), "processed/parsed_files/data_final.bim"
        )

        one_hot_outputs = self.input()
        return [luigi.LocalTarget(str(output_path)), one_hot_outputs]



import os
import subprocess
from dataclasses import fields, dataclass
from functools import lru_cache
from pathlib import Path
from statistics import mean
from tempfile import TemporaryDirectory
from typing import Dict, Any, Optional, Iterable, Literal, Sequence

import luigi
import pandas as pd
import psutil
import torch
import yaml
from aislib.misc_utils import ensure_path_exists
from eir.setup.config import recursive_dict_replace

from eir_auto_gp.modelling.configs import get_aggregate_config, AggregateConfig
from eir_auto_gp.modelling.dl_feature_selection import get_genotype_subset_snps_file
from eir_auto_gp.modelling.gwas_feature_selection import run_gwas_feature_selection
from eir_auto_gp.preprocess.converge import ParseDataWrapper
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class RunModellingWrapper(luigi.Task):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        task_object = (
            TestSingleRun if self.modelling_config["do_test"] else TrainSingleRun
        )
        for fold in _get_fold_iterator(folds=str(self.folds)):
            yield task_object(
                fold=fold,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
            )

    def output(self):
        return self.input()


def _get_fold_iterator(folds: str) -> Iterable:
    if "-" in folds:
        start, end = folds.split("-")
        return range(int(start), int(end) + 1)
    elif "," in folds:
        return [int(i) for i in folds.split(",")]
    else:
        return [int(folds)]


class TestSingleRun(luigi.Task):
    fold = luigi.IntParameter()

    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        base = {
            "data": ParseDataWrapper(data_config=self.data_config),
            "train_run": TrainSingleRun(
                fold=self.fold,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
            ),
        }

        return base

    def run(self):
        output_root = Path(self.output().path).parent
        ensure_path_exists(output_root, is_folder=True)

        injection_params = build_injection_params(
            fold=int(self.fold),
            folder_with_runs=output_root.parent,
            genotype_data_path=self.data_config["genotype_data_path"],
            data_input_dict=self.input()["data"][0],
            task="test",
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
        )

        injections = _get_all_dynamic_injections(
            injection_params=injection_params,
            genotype_data_path=self.data_config["genotype_data_path"],
        )

        base_aggregate_config = get_aggregate_config()
        with TemporaryDirectory() as temp_dir:
            temp_config_folder = Path(temp_dir)
            build_configs(
                injections=injections,
                aggregate_config_base=base_aggregate_config,
                output_folder=temp_config_folder,
            )

            train_run_folder = Path(self.input()["train_run"].path).parent
            base_predict_command = get_testing_string_from_config_folder(
                config_folder=temp_config_folder, train_run_folder=train_run_folder
            )

            base_command_split = base_predict_command.split()
            process_info = subprocess.run(
                base_command_split,
                env=dict(os.environ, EIR_SEED=str(self.fold)),
            )

            if process_info.returncode != 0:
                raise RuntimeError(
                    f"Testing failed with return code {process_info.returncode}"
                )
            testing_complete_file = Path(self.output().path)
            testing_complete_file.touch()

    def output(self):
        run_folder = Path(self.input()["train_run"].path).parent
        output_path = run_folder / "test_set_predictions/test_complete.txt"
        return luigi.LocalTarget(str(output_path))


def get_testing_string_from_config_folder(
    config_folder: Path, train_run_folder: Path
) -> str:
    base_string = "eirpredict"
    globals_string = " --global_configs "
    inputs_string = " --input_configs "
    output_string = " --output_configs "

    for file in config_folder.iterdir():
        if file.suffix == ".yaml":
            if "global" in file.stem:
                globals_string += " " + f"{str(file)}" + " "
            elif "input" in file.stem:
                inputs_string += " " + f"{str(file)}" + " "
            elif "output" in file.stem:
                output_string += " " + f"{str(file)}" + " "

    final_string = base_string + globals_string + inputs_string + output_string

    saved_models = list((train_run_folder / "saved_models").iterdir())
    assert len(saved_models) == 1, "Expected only one saved model."

    final_string += f" --model_path {saved_models[0]} --evaluate"

    test_output_folder = train_run_folder / "test_set_predictions"
    ensure_path_exists(path=test_output_folder, is_folder=True)

    final_string += f" --output_folder {test_output_folder}"

    return final_string


class TrainSingleRun(luigi.Task):
    fold = luigi.IntParameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        base = {"data": ParseDataWrapper(data_config=self.data_config)}
        for i in range(self.fold):
            base[f"train_{i}"] = TrainSingleRun(
                fold=i,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
            )
        return base

    def run(self):
        output_root = Path(self.output().path).parent
        ensure_path_exists(output_root, is_folder=True)

        injection_params = build_injection_params(
            fold=int(self.fold),
            folder_with_runs=output_root.parent,
            genotype_data_path=self.data_config["genotype_data_path"],
            data_input_dict=self.input()["data"][0],
            task="train",
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
        )

        injections = _get_all_dynamic_injections(
            injection_params=injection_params,
            genotype_data_path=self.data_config["genotype_data_path"],
        )

        base_aggregate_config = get_aggregate_config()
        with TemporaryDirectory() as temp_dir:
            temp_config_folder = Path(temp_dir)
            build_configs(
                injections=injections,
                aggregate_config_base=base_aggregate_config,
                output_folder=temp_config_folder,
            )

            base_train_command = get_training_string_from_config_folder(
                config_folder=temp_config_folder
            )

            base_command_split = base_train_command.split()
            process_info = subprocess.run(
                base_command_split,
                env=dict(os.environ, EIR_SEED=str(self.fold)),
            )

            if process_info.returncode != 0:
                raise RuntimeError(
                    f"Training failed with return code {process_info.returncode}"
                )
            training_complete_file = Path(self.output().path)
            training_complete_file.touch()

    def output(self):
        modelling_output_folder = Path(self.modelling_config["modelling_output_folder"])
        fold_output_folder = modelling_output_folder / f"fold_{self.fold}"
        file_name = fold_output_folder / "completed_train.txt"
        local_target = luigi.LocalTarget(path=str(file_name))
        return local_target


@dataclass
class ModelInjectionParams:
    fold: int
    output_folder: str
    batch_size: int
    genotype_input_source: str
    genotype_subset_snps_file: Optional[str]
    label_file_path: str
    input_cat_columns: list[str]
    input_con_columns: list[str]
    output_cat_columns: list[str]
    output_con_columns: list[str]
    compute_attributions: bool
    weighted_sampling_columns: list[str]


def build_injection_params(
    fold: int,
    folder_with_runs: Path,
    genotype_data_path: str,
    data_input_dict: Dict[str, luigi.LocalTarget],
    task: Literal["train", "test"],
    data_config: Dict[str, Any],
    feature_selection_config: Dict[str, Any],
    modelling_config: Dict[str, Any],
) -> ModelInjectionParams:
    compute_attributions = False
    n_act_folds = feature_selection_config["n_dl_feature_selection_folds"]
    if task == "train" and fold < n_act_folds:
        compute_attributions = True

    weighted_sampling_columns = None
    if modelling_config["output_cat_columns"]:
        weighted_sampling_columns = ["all"]

    feature_selection_tasks = feature_selection_config["feature_selection"]

    gwas_manual_subset = None
    if feature_selection_tasks is not None:
        if "gwas" in feature_selection_tasks:
            gwas_manual_subset = run_gwas_feature_selection(
                genotype_data_path=genotype_data_path,
                data_config=data_config,
                modelling_config=modelling_config,
                feature_selection_config=feature_selection_config,
            )

    bim_file = _get_bim_path(genotype_data_path=genotype_data_path)
    snp_subset_file = get_genotype_subset_snps_file(
        fold=fold,
        folder_with_runs=folder_with_runs,
        feature_selection_approach=feature_selection_config["feature_selection"],
        feature_selection_output_folder=Path(
            feature_selection_config["feature_selection_output_folder"]
        ),
        bim_file=bim_file,
        n_dl_feature_selection_folds=n_act_folds,
        manual_subset_from_gwas=gwas_manual_subset,
    )

    base_output_folder = modelling_config["modelling_output_folder"]
    cur_run_output_folder = f"{base_output_folder}/fold_{fold}"

    params = ModelInjectionParams(
        fold=fold,
        output_folder=cur_run_output_folder,
        batch_size=64,
        genotype_input_source=data_input_dict[f"{task}_genotype"].path,
        genotype_subset_snps_file=snp_subset_file,
        label_file_path=data_input_dict[f"{task}_tabular"].path,
        input_cat_columns=modelling_config["input_cat_columns"],
        input_con_columns=modelling_config["input_con_columns"],
        output_cat_columns=modelling_config["output_cat_columns"],
        output_con_columns=modelling_config["output_con_columns"],
        compute_attributions=compute_attributions,
        weighted_sampling_columns=weighted_sampling_columns,
    )

    return params


def build_configs(
    aggregate_config_base: AggregateConfig,
    injections: Dict[str, Any],
    output_folder: Path,
) -> None:
    for config_field in fields(aggregate_config_base):
        config_name = config_field.name
        config = getattr(aggregate_config_base, config_name)
        if config_name in injections:
            config = recursive_dict_replace(
                dict_=config, dict_to_inject=injections[config_name]
            )
        else:
            continue

        validate_complete_config(config_element=config)
        with open(output_folder / f"{config_name}.yaml", "w") as f:
            yaml.dump(config, f)


def validate_complete_config(config_element: dict | list | str) -> None:
    match config_element:
        case dict(config_element):
            for key, value in config_element.items():
                validate_complete_config(config_element=value)
        case list(config_element):
            for value in config_element:
                validate_complete_config(config_element=value)
        case str(config_element):
            assert config_element != "FILL"


def get_training_string_from_config_folder(config_folder: Path) -> str:
    base_string = "eirtrain"
    globals_string = " --global_configs "
    inputs_string = " --input_configs "
    output_string = " --output_configs "

    for file in config_folder.iterdir():
        if file.suffix == ".yaml" and "_test" not in file.stem:
            if "global" in file.stem:
                globals_string += " " + f"{str(file)}" + " "
            elif "input" in file.stem:
                inputs_string += " " + f"{str(file)}" + " "
            elif "output" in file.stem:
                output_string += " " + f"{str(file)}" + " "

    final_string = base_string + globals_string + inputs_string + output_string

    return final_string


def _get_global_injections(
    fold: int,
    output_folder: str,
    valid_size: int,
    n_snps: int,
    n_samples: int,
    compute_attributions: bool,
    iter_per_epoch: int,
    weighted_sampling_columns: list[str],
) -> Dict[str, Any]:
    mixing_candidates = [0.2]
    cur_mixing = mixing_candidates[fold % len(mixing_candidates)]

    device = _get_device()
    memory_dataset = _get_memory_dataset(n_snps=n_snps, n_samples=n_samples)
    n_workers = _get_dataloader_workers(memory_dataset=memory_dataset)
    early_stopping_buffer = min(5000, iter_per_epoch * 5)
    sample_interval = min(2000, iter_per_epoch)

    injections = {
        "output_folder": output_folder,
        "device": device,
        "valid_size": valid_size,
        "dataloader_workers": n_workers,
        "memory_dataset": memory_dataset,
        "mixing_alpha": cur_mixing,
        "sample_interval": sample_interval,
        "checkpoint_interval": sample_interval,
        "early_stopping_buffer": early_stopping_buffer,
        "compute_attributions": compute_attributions,
        "weighted_sampling_columns": weighted_sampling_columns,
    }

    return injections


def _get_memory_dataset(n_snps: int, n_samples: int) -> bool:
    available_memory = psutil.virtual_memory().available
    upper_bound = 0.6 * available_memory

    # 4 for one-hot encoding
    total_size = n_snps * n_samples * 4

    percent = total_size / available_memory
    if total_size < upper_bound:
        logger.info(
            "Estimated dataset size %.4f GB is %.4f%% of available memory %.4f GB, "
            "using memory dataset.",
            total_size / 1e9,
            percent * 100,
            available_memory / 1e9,
        )
        return True

    logger.info(
        "Estimated dataset size %.4f GB is %.4f%% of available memory %.4f GB, "
        "using disk dataset.",
        total_size / 1e9,
        percent * 100,
        available_memory / 1e9,
    )
    return False


def _get_device() -> str:
    device = "cuda:0" if torch.cuda.is_available() else "cpu"

    if device == "cpu":
        logger.warning(
            "Using CPU as no CUDA device found, "
            "this might be much slower than using a CUDA device."
        )
    elif device == "cuda:0":
        logger.info("Using CUDA device 0 for modelling.")

    return device


def _get_dataloader_workers(memory_dataset: bool) -> int:
    if memory_dataset:
        return 0

    n_cores = os.cpu_count()
    n_workers = int(0.8 * n_cores / 2)

    if n_workers <= 2:
        n_workers = 0

    logger.info(
        "Using %d workers for data loading based on %d available cores.",
        n_workers,
        n_cores,
    )
    return n_workers


def _get_genotype_injections(
    input_source: str,
    genotype_use_snps_file: Optional[str],
) -> Dict[str, Any]:
    base_snp_path = (
        Path(input_source).parent.parent / "processed/parsed_files/data_final.bim"
    )
    assert base_snp_path.exists(), f"SNP file not found at {base_snp_path}"

    injections = {
        "input_info": {
            "input_source": input_source,
        },
        "input_type_info": {
            "snp_file": str(base_snp_path),
        },
    }

    if genotype_use_snps_file is not None:
        injections["input_type_info"]["subset_snps_file"] = str(genotype_use_snps_file)

    return injections


def _get_tabular_injections(
    input_source: str, input_cat_columns: list[str], input_con_columns: list[str]
) -> Dict[str, Any]:
    injections = {
        "input_info": {
            "input_source": input_source,
        },
        "input_type_info": {
            "input_cat_columns": input_cat_columns,
            "input_con_columns": input_con_columns,
        },
    }
    return injections


def _get_output_injections(
    label_file_path: str, output_cat_columns: list[str], output_con_columns: list[str]
) -> Dict[str, Any]:
    injections = {
        "output_info": {
            "output_source": label_file_path,
        },
        "output_type_info": {
            "target_cat_columns": output_cat_columns,
            "target_con_columns": output_con_columns,
        },
    }

    return injections


def _get_all_dynamic_injections(
    injection_params: ModelInjectionParams, genotype_data_path: str
) -> Dict[str, Any]:
    mip = injection_params

    samples_per_epoch = get_samples_per_epoch(model_injection_params=mip)
    valid_size = get_dynamic_valid_size(
        num_samples_per_epoch=samples_per_epoch, batch_size=mip.batch_size
    )
    iter_per_epoch = get_num_iter_per_epoch(
        num_samples_per_epoch=samples_per_epoch,
        batch_size=mip.batch_size,
        valid_size=valid_size,
    )
    bim_path = _get_bim_path(genotype_data_path=genotype_data_path)

    n_snps = _lines_in_file(file_path=bim_path)
    if mip.genotype_subset_snps_file is not None:
        n_snps = _lines_in_file(file_path=mip.genotype_subset_snps_file)

    n_samples = _lines_in_file(file_path=mip.label_file_path) - 1

    injections = {
        "global_config": _get_global_injections(
            fold=mip.fold,
            output_folder=mip.output_folder,
            valid_size=valid_size,
            iter_per_epoch=iter_per_epoch,
            n_snps=n_snps,
            n_samples=n_samples,
            compute_attributions=mip.compute_attributions,
            weighted_sampling_columns=mip.weighted_sampling_columns,
        ),
        "input_genotype_config": _get_genotype_injections(
            input_source=mip.genotype_input_source,
            genotype_use_snps_file=mip.genotype_subset_snps_file,
        ),
        "output_config": _get_output_injections(
            label_file_path=mip.label_file_path,
            output_cat_columns=list(mip.output_cat_columns),
            output_con_columns=list(mip.output_con_columns),
        ),
    }

    if mip.input_cat_columns or mip.input_con_columns:
        injections["input_tabular_config"] = _get_tabular_injections(
            input_source=mip.label_file_path,
            input_cat_columns=list(mip.input_cat_columns),
            input_con_columns=list(mip.input_con_columns),
        )

    return injections


def get_dynamic_valid_size(
    num_samples_per_epoch: int, batch_size: int, valid_size_upper_bound: int = 20000
) -> int:
    valid_size = int(0.1 * num_samples_per_epoch)

    if valid_size < batch_size:
        valid_size = batch_size

    if valid_size > valid_size_upper_bound:
        valid_size = valid_size_upper_bound

    return valid_size


@lru_cache()
def _lines_in_file(file_path: str | Path) -> int:
    with open(file_path, "r") as f:
        num_lines = sum(1 for _ in f)
    return num_lines


def _get_bim_path(genotype_data_path: str) -> str:
    bim_files = [i for i in Path(genotype_data_path).glob("*.bim")]
    assert len(bim_files) == 1, bim_files

    path = bim_files[0]
    assert path.exists(), f".bim file not found at {path}"
    return str(path)


def get_samples_per_epoch(model_injection_params: ModelInjectionParams) -> int:
    mip = model_injection_params

    if not mip.weighted_sampling_columns:
        num_samples = _lines_in_file(file_path=mip.label_file_path) - 1
        return num_samples

    logger.info(
        "Setting up weighted sampling for categorical output columns: %s.",
        mip.output_cat_columns,
    )
    label_counts = get_column_label_counts(
        label_file_path=mip.label_file_path, output_cat_columns=mip.output_cat_columns
    )

    mean_per_target = (min(i.values()) for i in label_counts.values())
    mean_all_outputs = int(mean(mean_per_target))

    return mean_all_outputs


def get_column_label_counts(
    label_file_path: str | Path, output_cat_columns: Sequence[str]
) -> Dict[str, Dict[str, int]]:
    columns = ["ID"] + list(output_cat_columns)
    df = pd.read_csv(label_file_path, index_col=["ID"], usecols=columns)

    label_counts = {}

    for col in output_cat_columns:
        label_counts[col] = df[col].value_counts().to_dict()

    return label_counts


def get_num_iter_per_epoch(
    num_samples_per_epoch: int, batch_size: int, valid_size: int
) -> int:
    iter_per_epoch = (num_samples_per_epoch - valid_size) // batch_size
    iter_per_epoch = max(50, iter_per_epoch)
    return iter_per_epoch


if __name__ == "__main__":
    luigi.build([RunModellingWrapper()], local_scheduler=True)



import subprocess
from pathlib import Path
from typing import Dict, Any, Sequence, Optional

from aislib.misc_utils import ensure_path_exists, get_logger

from eir_auto_gp.preprocess import gwas_pre_selection as gps

logger = get_logger(name=__name__)


def get_gwas_feature_selection_filter_config(
    genotype_data_path: str,
    data_config: Dict[str, Any],
    modelling_config: Dict[str, Any],
    feature_selection_config: Dict[str, Any],
) -> gps.GWASPreFilterConfig:
    target_names = (
        modelling_config["output_con_columns"] + modelling_config["output_cat_columns"]
    )
    covariate_names = (
        modelling_config["input_con_columns"] + modelling_config["input_cat_columns"]
    )

    fam_file_path = next(Path(genotype_data_path).glob("*.fam"))
    id_split_folder = _get_id_split_folder(
        pre_split_folder=data_config["pre_split_folder"],
        data_output_folder=data_config["data_output_folder"],
        fam_file=fam_file_path,
    )

    config = gps.GWASPreFilterConfig(
        genotype_data_path=genotype_data_path,
        label_file_path=data_config["label_file_path"],
        output_path=feature_selection_config["feature_selection_output_folder"],
        target_names=target_names,
        covariate_names=covariate_names,
        pre_split_folder=id_split_folder,
        only_gwas=True,
        p_value_threshold=feature_selection_config["gwas_p_value_threshold"],
    )

    return config


def _get_id_split_folder(
    pre_split_folder: Optional[str],
    data_output_folder: str,
    fam_file: Optional[Path] = None,
) -> str:
    if pre_split_folder is not None and Path(pre_split_folder).exists():
        return pre_split_folder

    ids_folder = Path(data_output_folder, "ids")
    if ids_folder.exists():
        gps.add_plink_format_train_test_files(fam_file=fam_file, ids_folder=ids_folder)
        return str(ids_folder)

    raise ValueError(f"Could not find ids folder in {data_output_folder}")


def run_gwas_feature_selection(
    genotype_data_path: str,
    data_config: Dict[str, Any],
    modelling_config: Dict[str, Any],
    feature_selection_config: Dict[str, Any],
) -> str | Path:
    filter_config = get_gwas_feature_selection_filter_config(
        genotype_data_path=genotype_data_path,
        data_config=data_config,
        modelling_config=modelling_config,
        feature_selection_config=feature_selection_config,
    )

    fam_file_path = next(Path(genotype_data_path).glob("*.fam"))

    gwas_label_path = Path(filter_config.output_path, "gwas_label_file.csv")
    ensure_path_exists(path=gwas_label_path)
    if not gwas_label_path.exists():
        gps.prepare_gwas_label_file(
            label_file_path=filter_config.label_file_path,
            fam_file_path=fam_file_path,
            output_path=gwas_label_path,
        )

    base_path = fam_file_path.with_suffix("")
    fs_output_folder = feature_selection_config["feature_selection_output_folder"]
    gwas_output_path = Path(fs_output_folder, "gwas_output")

    train_ids_file = Path(filter_config.pre_split_folder, "train_ids_plink.txt")
    assert train_ids_file.exists(), f"Could not find train ids file at {train_ids_file}"

    command = gps.get_plink_gwas_command(
        base_path=base_path,
        label_file_path=gwas_label_path,
        target_names=filter_config.target_names,
        covariate_names=filter_config.covariate_names,
        output_path=gwas_output_path,
        ids_file=train_ids_file,
    )

    if not all_gwas_already_finished(
        target_names=filter_config.target_names,
        gwas_output_folder=gwas_output_path,
    ):
        logger.info("Running GWAS with command: %s", " ".join(command))
        subprocess.run(command, check=True)
        gps.plot_gwas_results(
            gwas_output_path=gwas_output_path,
            p_value_line=filter_config.p_value_threshold,
        )
        gwas_label_path.unlink()

    snps_to_keep_path = gps.gather_all_snps_to_keep(
        gwas_output_folder=gwas_output_path,
        p_value_threshold=filter_config.p_value_threshold,
    )

    return snps_to_keep_path


def all_gwas_already_finished(
    target_names: Sequence[str], gwas_output_folder: str | Path
) -> bool:
    for target_name in target_names:
        gwas_file = tuple(
            i for i in Path(gwas_output_folder).glob(f"gwas.{target_name}*")
        )
        if not gwas_file:
            return False

    return True



import json
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Iterator, Dict, Sequence, Literal, Tuple

import luigi
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from aislib.misc_utils import ensure_path_exists
from eir.experiment_io.experiment_io import load_serialized_train_experiment
from eir.train_utils.metrics import al_metric_record_dict
from eir.train_utils.train_handlers import _iterdir_ignore_hidden

from eir_auto_gp.modelling.run_modelling import RunModellingWrapper
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class RunAnalysisWrapper(luigi.WrapperTask):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()
    analysis_config = luigi.DictParameter()

    def requires(self):
        yield RunModellingWrapper(
            folds=self.folds,
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
        )
        yield GatherValidationResults(
            folds=self.folds,
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
            analysis_config=self.analysis_config,
        )
        yield GatherFeatureSelectionResults(
            folds=self.folds,
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
            analysis_config=self.analysis_config,
        )
        if self.modelling_config["do_test"]:
            yield GatherTestResults(
                folds=self.folds,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
                analysis_config=self.analysis_config,
            )


class GatherFeatureSelectionResults(luigi.Task):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()
    analysis_config = luigi.DictParameter()

    def requires(self):
        return {
            "runs": RunModellingWrapper(
                folds=self.folds,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
            ),
            "validation_results": GatherValidationResults(
                folds=self.folds,
                data_config=self.data_config,
                feature_selection_config=self.feature_selection_config,
                modelling_config=self.modelling_config,
                analysis_config=self.analysis_config,
            ),
        }

    def run(self):
        val_results = self.input()["validation_results"]
        feature_selection_folder = Path(
            self.feature_selection_config["feature_selection_output_folder"]
        )
        df_fractions = gather_fractions(
            feature_selection_folder=feature_selection_folder
        )
        for target_name, local_target in val_results.items():
            df_val_results = pd.read_csv(local_target.path, index_col=0)
            df_val_parsed = _parse_val_results(df=df_val_results)

            df_combined = df_val_parsed.join(df_fractions)

            output_folder = Path(self.analysis_config["analysis_output_folder"])
            ensure_path_exists(path=output_folder, is_folder=True)

            figure = _get_performance_fraction_figure(
                df=df_combined, target_name=target_name
            )
            output_path = output_folder / f"{target_name}_feature_selection.pdf"
            figure.savefig(output_path)

    def output(self):
        output_folder = Path(self.analysis_config["analysis_output_folder"])
        ensure_path_exists(path=output_folder, is_folder=True)

        targets = {}
        for target in chain(
            self.modelling_config["output_cat_columns"],
            self.modelling_config["output_con_columns"],
        ):
            targets[target] = luigi.LocalTarget(
                path=Path(output_folder, f"{target}_feature_selection.pdf")
            )

        return targets


def _parse_val_results(df: pd.DataFrame) -> pd.DataFrame:
    df_copy = df.copy()

    return df_copy[["Best Average Performance"]]


def _get_performance_fraction_figure(df: pd.DataFrame, target_name: str) -> plt.Figure:
    sns.set_theme(style="whitegrid")
    df_mean = df.groupby("Fraction SNPs").mean().reset_index()
    ax = sns.lineplot(
        data=df_mean,
        x="Fraction SNPs",
        y="Best Average Performance",
        color="black",
        linewidth=1,
    )

    ax.set_xscale("log")
    ax.set(xlabel="Fraction SNPs", ylabel="Average Val. Performance", title=target_name)

    return ax.figure


def gather_fractions(feature_selection_folder: Path) -> pd.DataFrame:
    fractions = []

    for file in _iterdir_ignore_hidden(
        path=feature_selection_folder / "dl_importance" / "snp_subsets"
    ):
        if "fraction" in file.stem:
            cur_fraction = float(file.read_text().strip())
            cur_fold = int(file.stem.split("_")[-1])
            fractions.append(
                {"Fold": f"Fold {cur_fold}", "Fraction SNPs": cur_fraction}
            )

    df_fractions = pd.DataFrame(data=fractions).set_index("Fold")

    return df_fractions


class GatherTestResults(luigi.Task):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()
    analysis_config = luigi.DictParameter()

    def requires(self):
        return RunModellingWrapper(
            folds=self.folds,
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
        )

    def run(self):
        output_folder = Path(self.analysis_config["analysis_output_folder"])
        ensure_path_exists(path=output_folder, is_folder=True)

        run_folder = Path(self.modelling_config["modelling_output_folder"])
        cat_targets = self.modelling_config["output_cat_columns"]
        con_targets = self.modelling_config["output_con_columns"]
        test_results_iterator = gather_test_predictions(
            folder_with_folds=run_folder,
            cat_targets=cat_targets,
            con_targets=con_targets,
        )

        for tr in test_results_iterator:
            tr.df_ensemble.to_csv(
                path_or_buf=output_folder / f"{tr.target_name}_test_predictions.csv",
                index=True,
            )
            tr.results.to_csv(
                path_or_buf=output_folder / f"{tr.target_name}_test_results.csv",
                index=True,
            )

    def output(self):
        analysis_output_folder = Path(self.analysis_config["analysis_output_folder"])
        aof = analysis_output_folder

        targets = {}
        for target in chain(
            self.modelling_config["output_cat_columns"],
            self.modelling_config["output_con_columns"],
        ):
            targets[f"{target}_test_results"] = luigi.LocalTarget(
                aof / f"{target}_test_results.csv"
            )
            targets[f"{target}_test_predictions"] = luigi.LocalTarget(
                aof / f"{target}_test_predictions.csv"
            )

        return targets


@dataclass
class TestResults:
    target_name: str
    df_ensemble: pd.DataFrame
    results: pd.DataFrame


def gather_test_predictions(
    folder_with_folds: Path, cat_targets: Sequence[str], con_targets: Sequence[str]
) -> Iterator[TestResults]:
    """
    TODO: Add mapping of numerical labels to original strings.
    """
    dfs = {}
    results = {}

    metric_funcs = None
    for fold in _iterdir_ignore_hidden(path=folder_with_folds):
        targets_folder = Path(fold, "test_set_predictions", "eir_auto_gp")

        if metric_funcs is None:
            logger.debug("Loading metric functions for ensemble from '%s'.", fold)
            metric_funcs = load_serialized_train_experiment(run_folder=fold).metrics

        for target_folder in _iterdir_ignore_hidden(path=targets_folder):
            target_name = target_folder.name

            if target_name not in dfs:
                dfs[target_name] = []

            if target_name not in results:
                results[target_name] = []

            predictions_file = Path(target_folder, "predictions.csv")
            df = pd.read_csv(filepath_or_buffer=predictions_file)
            dfs[target_name].append(df)

            cur_metrics = get_single_fold_results(fold=fold, target_name=target_name)
            results[target_name].append(cur_metrics)

    for target_name, dfs in dfs.items():
        df_ensemble = get_ensemble_predictions(dfs=dfs)

        target_type = get_target_type(
            target_name=target_name, cat_targets=cat_targets, con_targets=con_targets
        )
        ensemble_metrics = compute_metrics(
            df=df_ensemble,
            metrics=metric_funcs,
            target_type=target_type,
            target_name=target_name,
        )
        results[target_name].append(ensemble_metrics)

        df_results = pd.DataFrame(results[target_name])
        df_results = df_results.sort_values(by="Fold").set_index("Fold")

        yield TestResults(
            target_name=target_name,
            df_ensemble=df_ensemble,
            results=df_results,
        )


def get_ensemble_predictions(dfs: Sequence[pd.DataFrame]) -> pd.DataFrame:
    df_combined = pd.concat(dfs, axis=0)
    df_ensemble = df_combined.groupby(["ID", "True Label Untransformed"]).mean()
    df_ensemble = df_ensemble.reset_index().set_index("ID")

    return df_ensemble


def get_single_fold_results(fold: Path, target_name: str) -> Dict[str, float | str]:
    cur_metrics_file = Path(fold, "test_set_predictions", "calculated_metrics.json")
    cur_fold_metrics = json.load(open(cur_metrics_file, "r"))
    cur_metrics = cur_fold_metrics["eir_auto_gp"][target_name]

    parsed_metrics = {"Fold": fold.name.title().replace("_", " ")}

    for metric_name, metric_value in cur_metrics.items():
        parsed_metric_name = metric_name.split("_")[-1].upper()
        parsed_metrics[parsed_metric_name] = metric_value

    return parsed_metrics


def compute_metrics(
    df: pd.DataFrame,
    metrics: al_metric_record_dict,
    target_type: Literal["con", "cat"],
    target_name: str,
) -> Dict[str, float | str]:
    cur_metrics = metrics[target_type]

    metrics = {}
    for metric_record in cur_metrics:
        cur_labels = df["True Label"].values
        output_columns = [
            col
            for col in df.columns
            if "True Label" not in col and " Untransformed" not in col
        ]
        cur_outputs = df[output_columns].values

        cur_metric_func = metric_record.function

        cur_metric_value = cur_metric_func(
            labels=cur_labels,
            outputs=cur_outputs,
            output_name="eir_auto_gp",
            column_name=target_name,
        )

        cur_metric_name = metric_record.name.upper()
        metrics[cur_metric_name] = cur_metric_value

    metrics["Fold"] = "Ensemble"
    return metrics


def get_target_type(
    target_name: str, cat_targets: Sequence[str], con_targets: Sequence[str]
) -> Literal["con", "cat"]:
    if target_name in cat_targets:
        return "cat"
    elif target_name in con_targets:
        return "con"


class GatherValidationResults(luigi.Task):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    feature_selection_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()
    analysis_config = luigi.DictParameter()

    def requires(self):
        return RunModellingWrapper(
            folds=self.folds,
            data_config=self.data_config,
            feature_selection_config=self.feature_selection_config,
            modelling_config=self.modelling_config,
        )

    def run(self):
        output_folder = Path(self.analysis_config["analysis_output_folder"])
        ensure_path_exists(path=output_folder, is_folder=True)

        folder_with_folds = Path(self.modelling_config["modelling_output_folder"])

        val_results_iter = _gather_validation_results(
            folder_with_folds=folder_with_folds,
        )
        feature_selection_folder = Path(
            self.feature_selection_config["feature_selection_output_folder"]
        )
        df_fractions = gather_fractions(
            feature_selection_folder=feature_selection_folder
        )

        for target_name, df_results in val_results_iter:
            output_file = Path(output_folder, f"{target_name}_validation_results.csv")
            df_results = df_results.join(df_fractions, how="left")
            df_results.to_csv(path_or_buf=output_file)

    def output(self):
        output_folder = Path(self.analysis_config["analysis_output_folder"])
        ensure_path_exists(path=output_folder, is_folder=True)

        targets = {}
        for target in chain(
            self.modelling_config["output_cat_columns"],
            self.modelling_config["output_con_columns"],
        ):
            targets[target] = luigi.LocalTarget(
                path=Path(output_folder, f"{target}_validation_results.csv")
            )

        return targets


def _gather_validation_results(
    folder_with_folds: Path,
) -> Iterator[Tuple[str, pd.DataFrame]]:
    results = {}
    for fold_folder in _iterdir_ignore_hidden(path=folder_with_folds):
        targets_folder = Path(fold_folder, "results", "eir_auto_gp")

        avg_file = Path(fold_folder, "validation_average_history.log")
        df_average = pd.read_csv(filepath_or_buffer=avg_file)
        best_iter_idx = df_average["perf-average"].idxmax()
        best_val_average = df_average.loc[best_iter_idx]["perf-average"].item()

        for target_folder in _iterdir_ignore_hidden(path=targets_folder):
            target_name = target_folder.name

            if target_name not in results:
                results[target_name] = []

            val_file = Path(target_folder, f"validation_{target_name}_history.log")
            df_val_metrics = pd.read_csv(filepath_or_buffer=val_file)
            best_val_metrics = df_val_metrics.loc[best_iter_idx].to_dict()

            parsed_metrics = {
                "Fold": fold_folder.name.title().replace("_", " "),
                "Best Average Performance": best_val_average,
            }
            for metric_name, metric_value in best_val_metrics.items():
                parsed_metric_name = metric_name.split("_")[-1].upper()
                parsed_metrics[parsed_metric_name] = metric_value

            results[target_name].append(parsed_metrics)

    for target_name, results_list in results.items():
        df_results = pd.DataFrame(results_list)
        df_results = df_results.sort_values(by="Fold").set_index("Fold")

        yield target_name, df_results



import argparse
from argparse import RawTextHelpFormatter
import shutil
from copy import copy
from pathlib import Path
from typing import Dict, Any, Sequence

import luigi
import pandas as pd

from eir_auto_gp.analysis.run_analysis import RunAnalysisWrapper
from eir_auto_gp.utils.utils import get_logger
from eir_auto_gp.preprocess.gwas_pre_selection import validate_geno_data_path

logger = get_logger(name=__name__)


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        required=True,
        help="Root path to raw genotype data to be processed\n"
        "(e.g., containing my_data.bed, my_data.fam, my_data.bim).\n"
        "For this example, this parameter should be\n"
        "'/path/to/raw/genotype/data/'.\n"
        "Note that the file names are not included in this path,\n"
        "only the root folder. The file names are inferred, and\n"
        "*only one* set of files is expected.",
    )

    parser.add_argument(
        "--label_file_path",
        type=str,
        required=True,
        help="File path to label file with tabular inputs and labels to predict.",
    )

    parser.add_argument(
        "--global_output_folder",
        type=str,
        required=False,
        help="Common root folder to save data, feature selection and modelling results"
        " in.",
    )

    parser.add_argument(
        "--data_output_folder",
        type=str,
        required=False,
        help="Folder to save the processed data in.",
    )

    parser.add_argument(
        "--feature_selection_output_folder",
        type=str,
        required=False,
        help="Folder to save feature selection results in.",
    )

    parser.add_argument(
        "--modelling_output_folder",
        type=str,
        required=False,
        help="Folder to save modelling results in.",
    )

    parser.add_argument(
        "--analysis_output_folder",
        type=str,
        required=False,
        help="Folder to save analysis results in.",
    )

    parser.add_argument(
        "--output_name",
        type=str,
        default="genotype",
        help="Name used for dataset.",
    )

    parser.add_argument(
        "--pre_split_folder",
        type=str,
        required=False,
        help="If there is a pre-split folder, this will be used to\n"
        "split the data into train/val and test sets. If not,\n"
        "the data will be split randomly."
        "The folder should contain the following files:\n"
        "  - train.txt: List of sample IDs to use for training.\n"
        "  - test.txt: List of sample IDs to use for testing.\n"
        "If this option is not specified, the data will be split randomly"
        "into 90/10 train/test sets.",
    )

    parser.add_argument(
        "--feature_selection",
        default="gwas->dl",
        choices=["dl", "gwas", "gwas->dl", None],
        required=False,
        help="""\
    What kind of feature selection strategy to use for SNP selection:
      - If None, no feature selection is performed.
      - If 'dl', feature selection is performed using DL feature importance,
        and the top SNPs are selected iteratively using Bayesian optimization.
      - If 'gwas', feature selection is performed using GWAS p-values,
        as specified by the --gwas_p_value_threshold parameter.
      - If 'gwas->dl', feature selection is first performed using GWAS p-values,
        and then the top SNPs are selected iteratively using the DL importance method,
        but only on the SNPs under the GWAS threshold.
    """,
    )

    parser.add_argument(
        "--n_dl_feature_selection_folds",
        type=int,
        default=3,
        required=False,
        help="How many folds to run DL attribution calculation on genotype data\n"
        "before using results from attributions for feature selection.\n"
        "Applicable only if using 'dl' or 'gwas->dl' feature_selection options.",
    )

    parser.add_argument(
        "--gwas_p_value_threshold",
        type=float,
        required=False,
        default=1e-04,
        help="GWAS p-value threshold for filtering if using 'gwas' or 'gwas->dl'\n"
        "feature_selection options.",
    )

    parser.add_argument(
        "--folds",
        type=str,
        default="0-5",
        help="Training runs / folds to run, can be a single fold (e.g. 0),\n"
        "a range of folds (e.g. 0-5), or a comma-separated list of \n"
        "folds (e.g. 0,1,2,3,4,5).",
    )

    parser.add_argument(
        "--input_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as input.",
    )

    parser.add_argument(
        "--input_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as input.",
    )

    parser.add_argument(
        "--output_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as output.",
    )

    parser.add_argument(
        "--output_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as output.",
    )

    parser.add_argument(
        "--do_test",
        action="store_true",
        help="Whether to run test set prediction.",
    )

    return parser


def get_cl_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    cl_args = parser.parse_args()

    return cl_args


def validate_label_file(
    label_file_path: str,
    input_cat_columns: list[str],
    input_con_columns: list[str],
    output_cat_columns: list[str],
    output_con_columns: list[str],
) -> None:
    if not Path(label_file_path).exists():
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find {label_file_path}."
        )

    columns = pd.read_csv(label_file_path, nrows=1).columns
    if "ID" not in columns:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find 'ID' column."
        )

    all_columns = set(
        input_cat_columns + input_con_columns + output_cat_columns + output_con_columns
    )
    missing_columns = all_columns - set(columns)
    if len(missing_columns) > 0:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find columns {missing_columns}."
        )


def validate_targets(
    output_con_columns: list[str], output_cat_columns: list[str]
) -> None:
    if len(output_con_columns) == 0 and len(output_cat_columns) == 0:
        raise ValueError(
            "At least one output column must be specified as continuous or categorical."
        )

    if len(output_con_columns) + len(output_cat_columns) > 1:
        raise ValueError(
            "Currently only one target column per run is supported. Got "
            f"{output_con_columns} continuous and {output_cat_columns} "
            "categorical target columns."
        )


def validate_plink2_exists_in_path() -> None:
    if shutil.which("plink2") is None:
        raise ValueError(
            "plink2 is not installed or not in the path. "
            "Please install plink2 and try again."
        )


def run(cl_args: argparse.Namespace) -> None:
    validate_geno_data_path(geno_data_path=cl_args.genotype_data_path)
    validate_label_file(
        label_file_path=cl_args.label_file_path,
        input_cat_columns=cl_args.input_cat_columns,
        input_con_columns=cl_args.input_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
        output_con_columns=cl_args.output_con_columns,
    )
    validate_targets(
        output_con_columns=cl_args.output_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
    )
    validate_plink2_exists_in_path()

    cl_args = parse_output_folders(cl_args=cl_args)
    cl_args = _add_pre_split_folder_if_present(cl_args=cl_args)

    data_config = build_data_config(cl_args=cl_args)
    feature_selection_config = build_feature_selection_config(cl_args=cl_args)
    modelling_config = build_modelling_config(cl_args=cl_args)
    analysis_config = build_analysis_config(cl_args=cl_args)
    root_task = RunAnalysisWrapper(
        folds=cl_args.folds,
        data_config=data_config,
        feature_selection_config=feature_selection_config,
        modelling_config=modelling_config,
        analysis_config=analysis_config,
    )

    luigi.build(
        tasks=[root_task],
        workers=1,
        local_scheduler=True,
    )


def main():
    parser = get_argument_parser()
    cl_args = get_cl_args(parser=parser)

    run(cl_args=cl_args)


def parse_output_folders(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    if cl_args_copy.global_output_folder:
        gof = cl_args_copy.global_output_folder.rstrip("/")
        cl_args_copy.data_output_folder = gof + "/data"
        cl_args_copy.feature_selection_output_folder = gof + "/feature_selection"
        cl_args_copy.modelling_output_folder = gof + "/modelling"
        cl_args_copy.analysis_output_folder = gof + "/analysis"
    else:
        if not cl_args_copy.data_output_folder:
            raise ValueError(
                "Missing data output folder. "
                "Either a global output folder or a "
                "data output folder must be provided."
            )
        if not cl_args_copy.feature_selection_output_folder:
            raise ValueError(
                "Missing feature selection output folder. "
                "Either a global output folder or a "
                "feature selection output folder must be provided."
            )
        if not cl_args_copy.modelling_output_folder:
            raise ValueError(
                "Missing modelling output folder. "
                "Either a global output folder or a "
                "modelling output folder must be provided."
            )
        if not cl_args_copy.analysis_output_folder:
            raise ValueError(
                "Missing analysis output folder. "
                "Either a global output folder or a "
                "analysis output folder must be provided."
            )

    return cl_args_copy


def _add_pre_split_folder_if_present(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    genotype_path = Path(cl_args_copy.genotype_data_path)

    if (genotype_path / "ids").exists():
        cl_args_copy.pre_split_folder = str(genotype_path / "ids")
        logger.info(
            f"Found pre-split folder {cl_args_copy.pre_split_folder}. "
            f"in root genotype folder. Using those for train/test split."
        )

    return cl_args_copy


def build_data_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    data_keys = [
        "genotype_data_path",
        "label_file_path",
        "data_output_folder",
        "output_name",
        "pre_split_folder",
    ]

    base = extract_from_namespace(namespace=cl_args, keys=data_keys)
    base["output_format"] = "deeplake"

    return base


def build_feature_selection_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    feature_selection_keys = [
        "feature_selection_output_folder",
        "feature_selection",
        "n_dl_feature_selection_folds",
        "gwas_p_value_threshold",
    ]

    return extract_from_namespace(namespace=cl_args, keys=feature_selection_keys)


def build_modelling_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    modelling_keys = [
        "modelling_output_folder",
        "input_cat_columns",
        "input_con_columns",
        "output_cat_columns",
        "output_con_columns",
        "do_test",
    ]

    return extract_from_namespace(namespace=cl_args, keys=modelling_keys)


def build_analysis_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    analysis_keys = [
        "analysis_output_folder",
    ]

    return extract_from_namespace(namespace=cl_args, keys=analysis_keys)


def extract_from_namespace(
    namespace: argparse.Namespace, keys: Sequence[str]
) -> Dict[str, Any]:
    return {key: getattr(namespace, key) for key in keys}


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Any, Dict

from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


def get_base_global_config() -> Dict[str, Any]:
    base = {
        "output_folder": "FILL",
        "checkpoint_interval": "FILL",
        "batch_size": 64,
        "sample_interval": "FILL",
        "lr": 0.0002,
        "lr_plateau_patience": 4,
        "gradient_clipping": 1.0,
        "valid_size": "FILL",
        "n_epochs": 500,
        "dataloader_workers": "FILL",
        "device": "FILL",
        "early_stopping_buffer": "FILL",
        "early_stopping_patience": 6,
        "compute_attributions": False,
        "attribution_background_samples": 64,
        "max_attributions_per_class": 1000,
        "attributions_every_sample_factor": 5,
        "no_pbar": False,
        "mixing_alpha": "FILL",
        "optimizer": "adabelief",
    }
    return base


def get_base_input_genotype_config() -> Dict[str, Any]:
    base = {
        "input_info": {
            "input_source": "FILL",
            "input_name": "genotype",
            "input_type": "omics",
            "input_inner_key": "genotype",
        },
        "input_type_info": {
            "mixing_subtype": "cutmix-block",
            "na_augment_perc": 0.2,
            "na_augment_prob": 1.0,
            "snp_file": "FILL",
        },
        "model_config": {
            "model_type": "genome-local-net",
            "model_init_config": {
                "rb_do": 0.1,
                "channel_exp_base": 2,
                "kernel_width": 16,
                "first_kernel_expansion": -4,
                "l1": 0.0,
                "cutoff": 4096,
            },
        },
    }

    return base


def get_base_tabular_input_config() -> Dict[str, Any]:
    base = {
        "input_info": {
            "input_source": "FILL",
            "input_name": "eir_tabular",
            "input_type": "tabular",
        },
        "input_type_info": {
            "label_parsing_chunk_size": 20000,
            "input_cat_columns": ["FILL"],
            "input_con_columns": ["FILL"],
        },
        "model_config": {
            "model_type": "tabular",
            "model_init_config": {"fc_layer": True},
        },
    }

    return base


def get_base_fusion_config() -> Dict[str, Any]:
    base = {
        "model_config": {
            "fc_do": 0.1,
            "fc_task_dim": 512,
            "layers": [2],
            "rb_do": 0.1,
            "stochastic_depth_p": 0.1,
        },
        "model_type": "default",
    }
    return base


def get_base_output_config() -> Dict[str, Any]:
    base = {
        "output_info": {
            "output_name": "eir_auto_gp",
            "output_source": "FILL",
            "output_type": "tabular",
        },
        "output_type_info": {
            "target_con_columns": ["FILL"],
            "target_cat_columns": ["FILL"],
        },
        "model_config": {
            "model_type": "mlp_residual",
            "model_init_config": {
                "rb_do": 0.2,
                "fc_do": 0.2,
                "fc_task_dim": 512,
                "layers": [2],
                "stochastic_depth_p": 0.2,
                "final_layer_type": "linear",
            },
        },
    }
    return base


@dataclass(frozen=True)
class AggregateConfig:
    global_config: Dict[str, Any]
    input_genotype_config: Dict[str, Any]
    input_tabular_config: Dict[str, Any]
    fusion_config: Dict[str, Any]
    output_config: Dict[str, Any]


def get_aggregate_config() -> AggregateConfig:
    global_config = get_base_global_config()
    input_genotype_config = get_base_input_genotype_config()
    input_tabular_config = get_base_tabular_input_config()
    fusion_config = get_base_fusion_config()
    output_config = get_base_output_config()

    return AggregateConfig(
        global_config,
        input_genotype_config,
        input_tabular_config,
        fusion_config,
        output_config,
    )



from pathlib import Path
from shutil import copyfile
from typing import Sequence, Tuple, Dict, Optional
import warnings

warnings.filterwarnings("ignore", message=".*newer version of deeplake.*")

import deeplake
import eir.data_load.label_setup
import luigi
import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists
from eir.data_load.data_source_modules.deeplake_ops import load_deeplake_dataset

from eir_auto_gp.preprocess.genotype import FinalizeGenotypeParsing
from eir_auto_gp.preprocess.tabular import ParseLabelFile
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


@deeplake.compute
def _populate_deeplake_ds(
    id_array_tuple: Tuple[str, np.ndarray],
    deeplake_ds: deeplake.Dataset,
    output_name: str,
) -> None:
    """
    This would have been nice to have work with e.g. something like:

    with ds:
        parallel_function = _populate_deeplake_ds(
            output_name=output_name,
            argmax_replace_map=argmax_replace_map,
        )
        parallel_function.eval(data_in=id_array_generator, data_out=ds, num_workers=10)

    However, deeplake doesn't support having generators for eval yet.

    We could maybe do something like this for now, but needs to be tested:

    cur_chunk = []
    for index, (ids, arrays) in enumerate(id_array_generator):
        cur_chunk.append((ids, arrays))
        if len(cur_chunk) == parallel_chunk_size:
            with ds:
                parallel_function = _populate_deeplake_ds(
                    output_name=output_name,
                    argmax_replace_map=argmax_replace_map,
                )
                parallel_function.eval(
                    data_in=cur_chunk, data_out=ds, num_workers=10
                )

            cur_chunk = []

    if cur_chunk:
        with ds:
            parallel_function = _populate_deeplake_ds(
                output_name=output_name,
                argmax_replace_map=argmax_replace_map,
            )
            parallel_function.eval(data_in=cur_chunk, data_out=ds, num_workers=10)
    """

    id_, array = id_array_tuple

    sample = {
        "ID": id_,
        output_name: array,
    }

    deeplake_ds.append(sample)


class CommonSplitIntoTestSet(luigi.Task):
    genotype_data_path = luigi.Parameter()
    label_file_path = luigi.Parameter()
    data_output_folder = luigi.Parameter()
    output_format = luigi.Parameter()
    output_name = luigi.Parameter()
    pre_split_folder = luigi.Parameter()

    def requires(self):
        """
        Will require 3 tasks:
            - genotype processing,
            - label file processing.

        Will gather IDs from all of these, then generate a split from that.

        One option for configuration is to allow custom overwriting of QC command.
        """

        geno_task = FinalizeGenotypeParsing(
            raw_data_path=self.genotype_data_path,
            output_folder=str(self.data_output_folder) + "/genotype",
            output_name=self.output_name,
            output_format=self.output_format,
        )

        label_file_task = ParseLabelFile(
            output_folder=self.data_output_folder,
            label_file_path=self.label_file_path,
        )
        return {
            "genotype": geno_task,
            "label_file": label_file_task,
        }

    def run(self):
        inputs = self.input()

        genotype_path = Path(str(inputs["genotype"][1].path)) / "genotype"
        assert genotype_path.exists()
        label_file_path = Path(str(inputs["label_file"].path))

        output_root = Path(str(self.data_output_folder))

        common_ids_to_keep = _gather_all_ids(
            genotype_path=genotype_path, label_file=label_file_path
        )

        train_ids, test_ids = _id_setup_wrapper(
            output_root=output_root,
            pre_split_folder=self.pre_split_folder,
            common_ids_to_keep=common_ids_to_keep,
        )

        _split_csv_into_train_and_test(
            train_ids=train_ids,
            test_ids=test_ids,
            source=label_file_path,
            destination=output_root / "tabular" / "final",
        )

        _split_deeplake_ds_into_train_and_test(
            train_ids=train_ids,
            test_ids=test_ids,
            source=genotype_path,
            destination=output_root / "genotype" / "final",
        )

    def output(self):
        output_root = Path(str(self.data_output_folder))
        outputs = {}

        for split in ("train", "test"):
            cur_target = luigi.LocalTarget(str(output_root / "genotype/final" / split))
            outputs[f"{split}_genotype"] = cur_target

            cur_tabular_path = (
                output_root / "tabular/final" / f"labels_{split}"
            ).with_suffix(".csv")
            cur_tabular_target = luigi.LocalTarget(str(cur_tabular_path))
            outputs[f"{split}_tabular"] = cur_tabular_target

        return outputs


def _id_setup_wrapper(
    output_root: Path,
    common_ids_to_keep: Sequence[str],
    pre_split_folder: Optional[str] = None,
):
    if pre_split_folder is not None:
        pre_split_folder = Path(pre_split_folder)
        logger.info("Using pre-split folder: %s", pre_split_folder)

        train_path = pre_split_folder / "train_ids.txt"
        if not train_path.exists():
            raise FileNotFoundError(f"Could not find train IDs file: {train_path}")

        test_path = pre_split_folder / "test_ids.txt"
        if not test_path.exists():
            raise FileNotFoundError(f"Could not find test IDs file: {test_path}")

        train_ids = (
            pd.read_csv(train_path, header=None).astype(str).squeeze("columns").tolist()
        )
        test_ids = (
            pd.read_csv(test_path, header=None).astype(str).squeeze("columns").tolist()
        )

    else:
        logger.info("Generating train and test IDs.")

        train_ids, test_ids = _split_all_ids_into_train_and_test(
            all_ids=common_ids_to_keep
        )

        _save_ids_to_text_file(
            ids=train_ids, path=output_root / "ids" / "train_ids.txt"
        )
        _save_ids_to_text_file(ids=test_ids, path=output_root / "ids" / "test_ids.txt")

    logger.info("Train IDs: %d", len(train_ids))
    logger.info("Test IDs: %d", len(test_ids))
    return train_ids, test_ids


def _save_ids_to_text_file(ids: Sequence[str], path: Path) -> None:
    ensure_path_exists(path=path, is_folder=False)
    with open(path, "w") as f:
        for cur_id in ids:
            f.write(f"{cur_id}\n")


def _gather_all_ids(
    genotype_path: Path, label_file: Path, filter_common: bool = True
) -> Sequence[str]:
    genotype_ids = set(
        eir.data_load.label_setup.gather_ids_from_data_source(data_source=genotype_path)
    )
    label_file_ids = set(gather_ids_from_csv_file(file_path=label_file))

    all_ids = set().union(genotype_ids, label_file_ids)
    if filter_common:
        common_ids = label_file_ids.intersection(genotype_ids)
        logger.info(
            "Keeping %d common IDs among %d total (difference: %d).",
            len(common_ids),
            len(all_ids),
            len(all_ids) - len(common_ids),
        )
        return list(common_ids)

    return list(all_ids)


def gather_ids_from_csv_file(file_path: Path):
    logger.debug("Gathering IDs from %s.", file_path)
    df = pd.read_csv(file_path, usecols=["ID"])
    all_ids = tuple(df["ID"].astype(str))

    return all_ids


def _split_all_ids_into_train_and_test(
    all_ids: Sequence[str],
) -> Tuple[Sequence[str], Sequence[str]]:
    train_ids, test_ids = eir.data_load.label_setup.split_ids(
        ids=all_ids, valid_size=0.10
    )

    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    return train_ids, test_ids


def _split_deeplake_ds_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"
    if train_path.exists() and test_path.exists():
        return

    ensure_path_exists(path=train_path, is_folder=True)
    ensure_path_exists(path=test_path, is_folder=True)

    full_deeplake_ds = load_deeplake_dataset(data_source=str(source))

    ds_train = deeplake.empty(path=train_path)
    ds_test = deeplake.empty(path=test_path)

    for name, tensor in full_deeplake_ds.tensors.items():
        ds_train.create_tensor(
            name=name,
            htype=tensor.htype,
            dtype=tensor.dtype,
            sample_compression=tensor.meta.sample_compression,
        )
        ds_test.create_tensor(
            name=name,
            htype=tensor.htype,
            dtype=tensor.dtype,
            sample_compression=tensor.meta.sample_compression,
        )

    skipped_samples = 0
    with full_deeplake_ds, ds_train, ds_test:
        for idx, sample in enumerate(full_deeplake_ds):
            cur_id = sample["ID"].numpy().item()
            if cur_id in train_ids_set:
                parsed_sample = _deeplake_sample_to_dict(sample=sample)
                ds_train.append(parsed_sample)
            elif cur_id in test_ids_set:
                parsed_sample = _deeplake_sample_to_dict(sample=sample)
                ds_test.append(parsed_sample)
            else:
                skipped_samples += 1

            if idx % 10000 == 0:
                logger.info(
                    "Iterated over %d samples while splitting deeplake dataset into "
                    "train and test. Skipped %d samples.",
                    idx,
                    skipped_samples,
                )


def _deeplake_sample_to_dict(sample: deeplake.Dataset) -> Dict:
    """
    This is a workaround for the fact that deeplake does creates empty samples for
    *some* dtypes when we try to add them directly when iterating over samples.
    For example, numpy arrays are converted correctly, but strings result in empty
    for some reason (possibly a bug in deeplake).

    Note that this is possibly because deeplake .append is type to
    support Dict[str, Any], so maybe it is not a bug, but it just partially works
    when we feed it a sample.
    """

    parsed = {}

    for name, tensor in sample.tensors.items():
        if tensor.htype == "text":
            parsed[name] = tensor.numpy().item()
        else:
            parsed[name] = tensor.numpy()

    return parsed


def _split_arrays_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"
    if train_path.exists() and test_path.exists():
        return

    ensure_path_exists(path=train_path, is_folder=True)
    ensure_path_exists(path=test_path, is_folder=True)

    for array_file in source.iterdir():
        if array_file.stem in train_ids_set:
            copyfile(src=array_file, dst=train_path / array_file.name)
        elif array_file.stem in test_ids_set:
            copyfile(src=array_file, dst=test_path / array_file.name)


def _split_csv_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    ensure_path_exists(path=destination, is_folder=True)
    df_labels = pd.read_csv(filepath_or_buffer=source)
    df_labels["ID"] = df_labels["ID"].astype(str)

    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)

    df_train = df_labels[df_labels["ID"].isin(train_ids_set)].copy()
    df_test = df_labels[df_labels["ID"].isin(test_ids_set)].copy()

    df_train.to_csv(destination / "labels_train.csv", index=False)
    df_test.to_csv(destination / "labels_test.csv", index=False)


class ParseDataWrapper(luigi.Task):
    data_config = luigi.DictParameter()

    def requires(self):
        yield CommonSplitIntoTestSet(**self.data_config)

    def output(self):
        return self.input()



from pathlib import Path
from typing import Optional, Tuple, Iterator, Literal

import pandas as pd
from aislib.misc_utils import ensure_path_exists
from eir.train_utils.train_handlers import _iterdir_ignore_hidden
from eir.visualization.interpretation_visualization import plot_snp_manhattan_plots
from skopt import Optimizer

from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


def get_genotype_subset_snps_file(
    fold: int,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    bim_file: str | Path,
    feature_selection_approach: Literal["dl", "gwas", "gwas->dl", None],
    n_dl_feature_selection_folds: int,
    manual_subset_from_gwas: Optional[str | Path],
) -> Optional[Path]:
    match feature_selection_approach:
        case None:
            return None
        case "gwas":
            return manual_subset_from_gwas
        case "dl" | "gwas->dl":
            if feature_selection_approach == "gwas->dl":
                assert manual_subset_from_gwas is not None

            computed_subset_file = run_dl_bo_selection(
                fold=fold,
                folder_with_runs=folder_with_runs,
                feature_selection_output_folder=feature_selection_output_folder,
                bim_file=bim_file,
                n_dl_feature_selection_folds=n_dl_feature_selection_folds,
                manual_subset_from_gwas=manual_subset_from_gwas,
            )
            return computed_subset_file


def run_dl_bo_selection(
    fold: int,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    bim_file: str | Path,
    n_dl_feature_selection_folds: int,
    manual_subset_from_gwas: Optional[str | Path],
) -> Optional[Path]:
    fs_out_folder = feature_selection_output_folder
    subsets_out_folder = fs_out_folder / "dl_importance" / "snp_subsets"
    snp_subset_file = subsets_out_folder / f"dl_snps_{fold}.txt"
    if snp_subset_file.exists():
        return snp_subset_file

    fractions_file = subsets_out_folder / f"dl_snps_fraction_{fold}.txt"
    if fold < n_dl_feature_selection_folds:
        gwas_snps_or_none = _handle_dl_feature_selection_options(
            bim_file=bim_file,
            manual_subset_from_gwas_file=manual_subset_from_gwas,
            snp_subset_file=snp_subset_file,
            fractions_file=fractions_file,
        )
        return gwas_snps_or_none

    df_attributions = gather_eir_snp_attributions(folder_with_runs=folder_with_runs)

    importance_file = fs_out_folder / "dl_importance" / "dl_attributions.csv"
    ensure_path_exists(path=importance_file)
    df_attributions.to_csv(path_or_buf=importance_file)

    plot_snp_manhattan_plots(
        df_snp_grads=df_attributions,
        outfolder=importance_file.parent,
        title_extra="Aggregated",
    )

    top_n, fraction = get_auto_top_n(
        df_attributions=df_attributions,
        folder_with_runs=folder_with_runs,
        feature_selection_output_folder=feature_selection_output_folder,
        fold=fold,
    )
    logger.info("Top %d SNPs selected.", top_n)

    df_top_n = get_top_n_snp_list_df(df_attributions=df_attributions, top_n_snps=top_n)
    ensure_path_exists(path=snp_subset_file)
    df_top_n.to_csv(path_or_buf=snp_subset_file, index=False, header=False)
    fractions_file.write_text(str(fraction))

    return snp_subset_file


def _handle_dl_feature_selection_options(
    bim_file: str | Path,
    manual_subset_from_gwas_file: Optional[str | Path],
    snp_subset_file: Path,
    fractions_file: Path,
) -> Optional[Path]:
    ensure_path_exists(path=snp_subset_file)
    all_snps = _get_snps_in_bim_file(bim_file_path=bim_file)
    base_fraction = "1.0"

    if manual_subset_from_gwas_file is not None:
        logger.info("Using manual subset of SNPs.")
        gwas_snps = Path(manual_subset_from_gwas_file).read_text().splitlines()
        all_snps = gwas_snps

    snp_subset_file.write_text("\n".join(all_snps))

    ensure_path_exists(path=fractions_file)
    fractions_file.write_text(base_fraction)

    if manual_subset_from_gwas_file is not None:
        return snp_subset_file

    return None


def get_top_n_snp_list_df(
    df_attributions: pd.DataFrame, top_n_snps: int
) -> pd.DataFrame:
    target_columns = [i for i in df_attributions.columns if "Aggregated" in i]
    assert len(target_columns) == 1
    target_column = target_columns[0]

    df_top_n = df_attributions.nlargest(n=top_n_snps, columns=target_column)
    df_top_n = df_top_n.reset_index()
    df_top_n = df_top_n.rename(columns={"VAR_ID": "SNP"})
    df_top_n = df_top_n[["SNP"]]

    return df_top_n


def gather_eir_snp_attributions(folder_with_runs: Path) -> pd.DataFrame:
    """
    Currently this assumes that we have only computed the attributions on the full
    sets. Later we can add support for subset runs with attributions as well.
    """

    df_final = None
    gathered_series = {}
    counts = {}
    for fold in _iterdir_ignore_hidden(path=folder_with_runs):
        results_folder = fold / "results/eir_auto_gp"

        if not results_folder.exists():
            logger.warning("Results folder %s does not exist.", str(results_folder))
            continue

        for target_folder in _iterdir_ignore_hidden(path=results_folder):
            cur_act_iter = _get_attribution_iterator(target_folder=target_folder)

            for df_act, aggregate_name in cur_act_iter:
                if df_final is None:
                    df_final = df_act.drop(columns=[aggregate_name])

                series_act = df_act[aggregate_name]

                if aggregate_name not in gathered_series:
                    gathered_series[aggregate_name] = series_act
                    counts[aggregate_name] = 1
                else:
                    gathered_series[aggregate_name] += series_act
                    counts[aggregate_name] += 1

    for aggregate_name, series_act in gathered_series.items():
        cur_average = series_act / counts[aggregate_name]
        df_final[aggregate_name] = cur_average

    return df_final


def _get_attribution_iterator(
    target_folder: Path,
) -> Iterator[tuple[pd.DataFrame, str]]:
    for p in target_folder.rglob("*"):
        if p.name == "snp_attributions.csv":
            cur_csv = p

            df_cur = pd.read_csv(filepath_or_buffer=cur_csv)
            df_cur = df_cur.set_index("VAR_ID")
            df_cur = df_cur.drop(columns=["Unnamed: 0"])

            act_cols = [i for i in df_cur.columns if "_attributions" in i]

            aggregate_column = f"Aggregated {target_folder.name}_attributions"
            df_cur[aggregate_column] = abs(df_cur[act_cols]).sum(axis=1)

            df_cur = df_cur.drop(columns=act_cols)

            yield df_cur, aggregate_column


def _get_snps_in_bim_file(bim_file_path: str | Path) -> list[str]:
    with open(bim_file_path, "r") as f:
        snps = [line.split()[1] for line in f]
    return snps


def get_auto_top_n(
    df_attributions: pd.DataFrame,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    fold: int,
) -> Tuple[int, float]:
    manual_fractions = _get_manual_fractions(
        total_n_snps=len(df_attributions), min_snps_cutoff=16
    )

    if fold < len(manual_fractions):
        next_fraction = manual_fractions[fold]
    else:
        opt = Optimizer(dimensions=[(0.0, 1.0)])
        df_history = _gather_fractions_and_performances(
            folder_with_runs=folder_with_runs,
            feature_selection_output_folder=feature_selection_output_folder,
        )

        for t in df_history.itertuples():
            opt.tell([t.fraction], t.best_val_performance)

        next_fraction = opt.ask()[0]

    top_n = int(next_fraction * len(df_attributions))
    if top_n < 16:
        top_n = 16
        next_fraction = top_n / len(df_attributions)

    return top_n, next_fraction


def _get_manual_fractions(total_n_snps: int, min_snps_cutoff: int) -> list[float]:
    base = [0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 0.2]

    manual_fractions = []
    for i in base:
        cur = i * total_n_snps
        if cur >= min_snps_cutoff:
            manual_fractions.append(i)

    return manual_fractions


def _gather_fractions_and_performances(
    folder_with_runs: Path, feature_selection_output_folder: Path
) -> pd.DataFrame:
    df_val_performances = _gather_best_val_performances(
        folder_with_runs=folder_with_runs
    )
    df_snp_fractions = _get_snp_fractions(
        feature_selection_output_folder=feature_selection_output_folder
    )

    df = df_val_performances.merge(df_snp_fractions, left_index=True, right_index=True)

    df = df.sort_index()

    return df


def _gather_best_val_performances(folder_with_runs: Path) -> pd.DataFrame:
    results = {}
    for fold in _iterdir_ignore_hidden(path=folder_with_runs):
        if not (fold / "completed_train.txt").exists():
            continue

        df_val = pd.read_csv(filepath_or_buffer=fold / "validation_average_history.log")
        best_performance = df_val["perf-average"].max()
        fold_as_int = int(fold.stem.split("_")[-1])
        results[fold_as_int] = best_performance

    df = pd.DataFrame.from_dict(
        results, orient="index", columns=["best_val_performance"]
    )
    return df


def _get_snp_fractions(feature_selection_output_folder: Path) -> pd.DataFrame:
    results = {}
    for f in _iterdir_ignore_hidden(
        path=feature_selection_output_folder / "dl_importance" / "snp_subsets"
    ):
        if "_fraction" not in f.name:
            continue

        fold_as_int = int(f.stem.split("_")[-1])
        fraction = f.read_text()
        results[fold_as_int] = float(fraction)

    df = pd.DataFrame.from_dict(results, orient="index", columns=["fraction"])
    return df



import argparse
import subprocess
from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from qmplot import manhattanplot, qqplot
from qmplot.modules._qq import ppoints
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.preprocess.converge import (
    gather_ids_from_csv_file,
    _id_setup_wrapper,
)
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


def main():
    parser = get_gwas_parser()
    cl_args = parser.parse_args()
    validate_geno_data_path(geno_data_path=cl_args.genotype_data_path)

    filter_config_ = get_gwas_pre_filter_config(cl_args=cl_args)
    run_gwas_pre_filter_wrapper(filter_config=filter_config_)


def run_gwas_pre_filter_wrapper(filter_config: "GWASPreFilterConfig") -> None:
    fam_file_path = next(Path(filter_config.genotype_data_path).glob("*.fam"))

    gwas_label_path = Path(filter_config.output_path, "gwas_label_file.csv")
    prepare_gwas_label_file(
        label_file_path=filter_config.label_file_path,
        fam_file_path=fam_file_path,
        output_path=gwas_label_path,
    )

    common_ids_to_keep = gather_all_ids(
        fam_file_path=fam_file_path,
        label_file_path=filter_config.label_file_path,
    )

    _id_setup_wrapper(
        common_ids_to_keep=common_ids_to_keep,
        output_root=Path(filter_config.output_path),
        pre_split_folder=filter_config.pre_split_folder,
    )
    train_ids_plink, _ = add_plink_format_train_test_files(
        fam_file=fam_file_path,
        ids_folder=Path(filter_config.output_path, "ids"),
    )

    base_path = fam_file_path.with_suffix("")
    gwas_output_path = Path(filter_config.output_path, "gwas_output")
    command = get_plink_gwas_command(
        base_path=base_path,
        label_file_path=gwas_label_path,
        target_names=filter_config.target_names,
        covariate_names=filter_config.covariate_names,
        output_path=gwas_output_path,
        ids_file=train_ids_plink,
    )

    logger.info("Running GWAS with command: %s", " ".join(command))
    subprocess.run(command, check=True)
    plot_gwas_results(
        gwas_output_path=gwas_output_path, p_value_line=filter_config.p_value_threshold
    )
    gwas_label_path.unlink()

    snps_to_keep_path = gather_all_snps_to_keep(
        gwas_output_folder=gwas_output_path,
        p_value_threshold=filter_config.p_value_threshold,
    )

    if filter_config.only_gwas:
        logger.info("Only running GWAS, not filtering genotype data.")
        return

    filter_output_path = Path(filter_config.output_path, fam_file_path.stem)
    filter_command = _get_plink_filter_snps_command(
        base_path=base_path,
        snps_to_keep_path=snps_to_keep_path,
        output_path=filter_output_path,
    )

    logger.info("Running GWAS filter with command: %s", " ".join(filter_command))
    subprocess.run(filter_command, check=True)


def _get_plink_filter_snps_command(
    base_path: str | Path,
    snps_to_keep_path: str | Path,
    output_path: str | Path,
) -> list[str]:
    command_base = (
        f"plink2"
        f" --bfile {base_path}"
        f" --extract {snps_to_keep_path}"
        f" --make-bed"
        f" --out {output_path}"
    )

    return command_base.split()


def get_plink_gwas_command(
    base_path: str | Path,
    label_file_path: str | Path,
    target_names: list[str],
    covariate_names: list[str],
    output_path: str | Path,
    ids_file: Optional[str | Path],
) -> list[str]:
    ensure_path_exists(path=output_path, is_folder=True)

    pheno_names = get_pheno_names(
        label_file_path=label_file_path,
        target_names=target_names,
        covariate_names=covariate_names,
    )

    command_base = (
        f"plink2"
        f" --bfile {base_path}"
        " --1"
        f" --pheno {label_file_path}"
        f" --pheno-name {' '.join(pheno_names)}"
        f" --glm skip-invalid-pheno "
        f"firth-fallback hide-covar omit-ref no-x-sex allow-no-covars"
        f" --out {output_path}/gwas"
    )

    if covariate_names:
        command_base += f" --covar {label_file_path}"
        command_base += f" --covar-name {' '.join(covariate_names)}"
        command_base += " --covar-variance-standardize"

    if ids_file is not None:
        command_base += f" --keep {ids_file}"

    return command_base.split()


def get_pheno_names(
    label_file_path: Path, target_names: list[str], covariate_names: list[str]
) -> list[str]:
    if target_names:
        return target_names

    id_columns = ["ID", "FID", "IID"]
    all_columns = pd.read_csv(label_file_path, nrows=1, sep=r"\s+").columns.tolist()
    to_skip = id_columns + covariate_names
    inferred_target_names = [col for col in all_columns if col not in to_skip]

    logger.info(
        "No phenotype target names provided, "
        "inferring target names from label file: %s",
        inferred_target_names,
    )

    return inferred_target_names


def add_plink_format_train_test_files(
    fam_file: Path, ids_folder: Path
) -> tuple[Path, Path]:
    df_fam = _read_fam(fam_path=fam_file)
    df_fam = df_fam[[0, 1]]

    train_ids = (
        pd.read_csv(Path(ids_folder, "train_ids.txt"), header=None)[0]
        .astype(str)
        .tolist()
    )
    train_ids = set(train_ids)

    test_ids = (
        pd.read_csv(Path(ids_folder, "test_ids.txt"), header=None)[0]
        .astype(str)
        .tolist()
    )
    test_ids = set(test_ids)

    df_fam_train = df_fam[df_fam[1].isin(train_ids)]
    df_fam_test = df_fam[df_fam[1].isin(test_ids)]

    train_output_path = Path(ids_folder, "train_ids_plink.txt")
    test_output_path = Path(ids_folder, "test_ids_plink.txt")

    df_fam_train.to_csv(train_output_path, sep="\t", header=False, index=False)
    df_fam_test.to_csv(test_output_path, sep="\t", header=False, index=False)

    return train_output_path, test_output_path


def _read_fam(fam_path: Path) -> pd.DataFrame:
    df_fam = pd.read_csv(
        fam_path,
        sep=r"\s+",
        header=None,
        dtype={0: str, 1: str},
    )

    return df_fam


def gather_all_snps_to_keep(gwas_output_folder: Path, p_value_threshold: float) -> Path:
    snps_to_keep = set()
    for gwas_file in gwas_output_folder.iterdir():
        if "logistic" not in gwas_file.name and "linear" not in gwas_file.name:
            continue

        cur_snps = _gather_snps_to_keep_from_gwas_output(
            gwas_file_path=gwas_file,
            p_value_threshold=p_value_threshold,
        )
        snps_to_keep.update(cur_snps)

    snps_to_keep = list(snps_to_keep)
    logger.info("Keeping %d SNPs in total.", len(snps_to_keep))

    snps_to_keep_path = Path(gwas_output_folder, "snps_to_keep.txt")
    with open(snps_to_keep_path, "w") as f:
        f.write("\n".join(snps_to_keep))

    return snps_to_keep_path


def _gather_snps_to_keep_from_gwas_output(
    gwas_file_path: str | Path,
    p_value_threshold: float,
) -> list[str]:
    df_gwas = pd.read_csv(gwas_file_path, sep="\t")

    snps_to_keep = df_gwas[df_gwas["P"] < p_value_threshold]["ID"].tolist()
    logger.info(
        "Keeping %d SNPs from %s with p-value < %f (total: %d).",
        len(snps_to_keep),
        gwas_file_path,
        p_value_threshold,
        len(df_gwas),
    )

    return snps_to_keep


def gather_all_ids(fam_file_path: str | Path, label_file_path: str | Path) -> list[str]:
    df_fam = _read_fam(fam_path=fam_file_path)
    genotype_ids = set(df_fam[1].astype(str))

    labelled_ids = set(gather_ids_from_csv_file(file_path=label_file_path))

    common_ids_to_keep = set().union(genotype_ids, labelled_ids)
    common_ids = genotype_ids.intersection(labelled_ids)

    logger.info(
        "Keeping %d common IDs among %d total (difference: %d).",
        len(common_ids),
        len(common_ids_to_keep),
        len(common_ids_to_keep) - len(common_ids),
    )

    return list(common_ids)


def _get_train_ids_file(filter_config: "GWASPreFilterConfig") -> Path:
    if filter_config.pre_split_folder:
        train_ids_file = Path(filter_config.pre_split_folder, "train_ids.txt")
    else:
        train_ids_file = Path(filter_config.output_path, "ids/train_ids.txt")
    assert train_ids_file.exists(), f"Train ids file {train_ids_file} does not exist."

    return train_ids_file


def prepare_gwas_label_file(
    label_file_path: str | Path, fam_file_path: str | Path, output_path: str | Path
) -> Path:
    df = pd.read_csv(filepath_or_buffer=label_file_path, index_col=0, dtype={"ID": str})

    df_fam = _read_fam(fam_path=fam_file_path)

    iid_to_fid = dict(zip(df_fam[1], df_fam[0]))

    df.insert(0, "FID", df.index.map(iid_to_fid))
    df.insert(1, "IID", df.index)

    df = _prepare_df_columns_for_gwas(df=df)

    df.to_csv(path_or_buf=output_path, sep="\t", index=False)

    return Path(output_path)


def _prepare_df_columns_for_gwas(df: pd.DataFrame) -> pd.DataFrame:
    """
    Note: We replace spaces with underscores in column names due to plink2 assuming
    tab/space separated columns.

    Note: For now we just have some simple heuristics for determining which columns
    to one-hot encode. We can make this more sophisticated/configurable later.
    """
    one_hot_target_columns = []
    for column in df.columns:
        if column in ["FID", "IID"]:
            continue

        if df[column].dtype in ("object", "category"):
            one_hot_target_columns.append(column)

        elif df[column].dtype == "int":
            n_unique = df[column].nunique()
            if 2 < n_unique < 10:
                one_hot_target_columns.append(column)
            else:
                logger.info(
                    "Column %s has %d unique values, not one-hot encoding.",
                    column,
                    n_unique,
                )

    df = pd.get_dummies(df, columns=one_hot_target_columns)

    df.columns = df.columns.str.replace(" ", "_")
    df = df.fillna(-9)

    return df


def plot_gwas_results(
    gwas_output_path: Path, p_value_line: Optional[float] = None
) -> None:
    for f in gwas_output_path.iterdir():
        if "linear" not in f.name and "logistic" not in f.name:
            continue

        df = pd.read_csv(f, sep="\t")
        df = df.dropna(how="any", axis=0)

        fig_manhattan = get_manhattan_plot(df=df, p_value_line=p_value_line)
        manhattan_output_path = Path(
            gwas_output_path, "plots", f"{f.name}_manhattan.png"
        )
        ensure_path_exists(path=manhattan_output_path)
        fig_manhattan.savefig(fname=manhattan_output_path, dpi=300)

        fig_qq = get_qq_plot(df=df)
        qq_output_path = Path(gwas_output_path, "plots", f"{f.name}_qq.png")
        ensure_path_exists(path=qq_output_path)
        fig_qq.savefig(fname=qq_output_path, dpi=300)


def get_manhattan_plot(df: pd.DataFrame, p_value_line: Optional[float]) -> plt.Figure:
    f, ax = plt.subplots(figsize=(12, 4), facecolor="w", edgecolor="k")
    ax = manhattanplot(
        data=df,
        marker=".",
        suggestiveline=p_value_line,
        sign_marker_p=None,
        sign_marker_color="r",
        snp="ID",
        title="",
        xlabel="Chromosome",
        ylabel=r"$-log_{10}{(P)}$",
        sign_line_cols=["#D62728", "#2CA02C"],
        hline_kws={"linestyle": "--", "lw": 1.3},
        is_annotate_topsnp=False,
        ld_block_size=500000,
        text_kws={
            "fontsize": 10,
            "arrowprops": dict(arrowstyle="-", color="k", alpha=0.6),
        },
        xticklabel_kws={"rotation": "vertical"},
        ax=ax,
    )

    fig = ax.get_figure()

    return fig


def get_qq_plot(df: pd.DataFrame) -> plt.Figure:
    f, ax = plt.subplots(figsize=(6, 6), facecolor="w", edgecolor="k")
    ax = qqplot(
        data=df["P"],
        marker="o",
        title="",
        xlabel=r"Expected $-log_{10}{(P)}$",
        ylabel=r"Observed $-log_{10}{(P)}$",
        ablinecolor="red",
        ax=ax,
    )

    data = df["P"]

    e = ppoints(len(data))
    o = -np.log10(sorted(data))
    e = -np.log10(e)

    ax.plot(
        [e.min(), o.min()],
        [e.max(), o.max()],
        color="red",
        linestyle="-",
    )

    fig = ax.get_figure()

    return fig


@dataclass
class GWASPreFilterConfig:
    genotype_data_path: str
    label_file_path: str
    output_path: str
    target_names: list[str]
    covariate_names: list[str]
    pre_split_folder: Optional[str]
    only_gwas: bool
    p_value_threshold: float = 1e-04


def get_gwas_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        required=True,
        help="Root path to raw genotype data to be processed"
        "(e.g., containing my_data.bed, my_data.fam, my_data.bim). "
        "For this example, this parameter should be "
        "'/path/to/data/raw/genotype/'."
        "Note that the file names is not included in this path, only the root folder."
        "The file names are inferred, and *only one* set of files is expected.",
    )

    parser.add_argument(
        "--label_file_path",
        type=str,
        help="Path to label file.",
    )

    parser.add_argument(
        "--output_path",
        type=str,
        help="Path to output folder.",
    )

    parser.add_argument(
        "--target_names",
        type=str,
        nargs="+",
        default=[],
        help="Which target columns to perform GWAS on.",
    )

    parser.add_argument(
        "--covariate_names",
        type=str,
        nargs="+",
        default=[],
        help="Covariate names to include in GWAS.",
    )

    parser.add_argument(
        "--pre_split_folder",
        type=str,
        required=False,
        help="Path to folder containing train.txt and test.txt.",
    )

    parser.add_argument(
        "--gwas_p_value_threshold",
        type=float,
        required=False,
        default=1e-04,
        help="GWAS p-value threshold for filtering.",
    )

    parser.add_argument(
        "--only_gwas",
        action="store_true",
        help="Only perform GWAS and do not filter genotype data.",
    )

    return parser


def get_gwas_pre_filter_config(cl_args: argparse.Namespace) -> GWASPreFilterConfig:
    return GWASPreFilterConfig(
        genotype_data_path=cl_args.genotype_data_path,
        label_file_path=cl_args.label_file_path,
        output_path=cl_args.output_path,
        target_names=cl_args.target_names,
        covariate_names=cl_args.covariate_names,
        pre_split_folder=cl_args.pre_split_folder,
        p_value_threshold=cl_args.gwas_p_value_threshold,
        only_gwas=cl_args.only_gwas,
    )


if __name__ == "__main__":
    main()


def validate_geno_data_path(geno_data_path: str) -> None:
    for suffix in [".bed", ".fam", ".bim"]:
        files = list(Path(geno_data_path).glob(f"*{suffix}"))
        if len(files) != 1:
            raise ValueError(
                f"Genotype data path {geno_data_path} is invalid. "
                f"Expected to find *exactly one* file with suffix {suffix}. "
                f"Found {files} ({len(files)} files) instead."
            )
